/*
 * Copyright 2012 Clemens Akens
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.clebert.jlingua;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import net.jcip.annotations.Immutable;

import com.google.common.collect.ImmutableSet;

/**
 * This class provides static methods for validating a context-free grammar.
 *
 * @author Clemens Akens
 */
@Immutable
public final class Validator {

    /**
     * This static method validates the specified context-free grammar against
     * the specified properties.
     *
     * @param grammar
     *            The context-free grammar to be validated.
     * @param properties
     *            The properties to be validated against.
     * @return The successfully validated context-free grammar.
     * @throws ValidationException
     *             Signals that the specified context-free grammar is invalid
     *             against the specified properties.
     */
    public static Grammar validate(final Grammar grammar, final Collection<Property> properties) throws ValidationException {
        return Validator.validate(grammar, properties.toArray(new Property[properties.size()]));
    }

    /**
     * This static method validates the specified context-free grammar against
     * the specified properties.
     *
     * @param grammar
     *            The context-free grammar to be validated.
     * @param properties
     *            The properties to be validated against.
     * @return The successfully validated context-free grammar.
     * @throws ValidationException
     *             Signals that the specified context-free grammar is invalid
     *             against the specified properties.
     */
    public static Grammar validate(final Grammar grammar, final Property... properties) throws ValidationException {
        for (final Property property : properties) {
            final List<String> errors = new ArrayList<String>();

            if (property.equals(Property.Clean)) {
                Validator.validateIsClean(grammar, errors);
            } else if (property.equals(Property.Disjoint)) {
                Validator.validateIsDisjoint(grammar, errors);
            } else if (property.equals(Property.EmptyFree)) {
                Validator.validateIsEmptyFree(grammar, errors);
            } else if (property.equals(Property.LeftRecursionFree)) {
                Validator.validateIsLeftRecursionFree(grammar, errors);
            } else if (property.equals(Property.UnitFree)) {
                Validator.validateIsUnitFree(grammar, errors);
            }

            if (errors.size() > 0) {
                final StringBuilder builder = new StringBuilder();

                builder.append("The context-free grammar contains ");
                builder.append(property.getDescription());
                builder.append(":");

                for (int i = 0; i < errors.size(); ++i) {
                    if (i > 0) {
                        builder.append(",");
                    }

                    builder.append(" ");
                    builder.append(errors.get(i));
                }

                throw new ValidationException(builder.toString());
            }
        }

        return grammar;
    }

    /**
     * This static method validates that the specified context-free grammar does
     * not contain non-productive production rules or unreachable non-terminal
     * symbols.
     *
     * @param grammar
     *            The context-free grammar to be validated.
     * @param errors
     *            The list to which errors that may occur can be added.
     */
    private static void validateIsClean(final Grammar grammar, final List<String> errors) {
        final Set<Rule> nonProductiveRules = new HashSet<Rule>(grammar.getRules());

        nonProductiveRules.removeAll(Utils.computeProductiveRulesByNonTerminal(grammar).values());

        for (final Rule rule : nonProductiveRules) {
            errors.add(rule.toObject().toString());
        }

        final Set<NonTerminal> unreachableNonTerminals = new HashSet<NonTerminal>(grammar.getNonTerminals());

        unreachableNonTerminals.removeAll(Utils.computeReachableNonTerminals(grammar));

        for (final NonTerminal nonTerminal : unreachableNonTerminals) {
            errors.add(nonTerminal.toObject().toString());
        }
    }

    /**
     * This static method validates that the specified context-free grammar does
     * not contain terminal symbols with the same string representation as
     * non-terminal symbols.
     *
     * @param grammar
     *            The context-free grammar to be validated.
     * @param errors
     *            The list to which errors that may occur can be added.
     */
    private static void validateIsDisjoint(final Grammar grammar, final List<String> errors) {
        final Set<String> nonTerminalNames = new HashSet<String>();

        for (final NonTerminal nonTerminal : grammar.getNonTerminals()) {
            nonTerminalNames.add(nonTerminal.getName());
        }

        for (final Terminal terminal : grammar.getTerminals()) {
            if (nonTerminalNames.contains(terminal.getName())) {
                errors.add(terminal.toObject().toString());
            }
        }
    }

    /**
     * This static method validates that the specified context-free grammar does
     * not contain nullable non-terminal symbols. Only the start symbol can be
     * nullable.
     *
     * @param grammar
     *            The context-free grammar to be validated.
     * @param errors
     *            The list to which errors that may occur can be added.
     */
    private static void validateIsEmptyFree(final Grammar grammar, final List<String> errors) {
        final ImmutableSet<NonTerminal> nullableNonTerminals = Utils.computeNullableNonTerminals(grammar);

        for (final NonTerminal nonTerminal : nullableNonTerminals) {
            if (!nonTerminal.equals(grammar.getStartSymbol())) {
                errors.add(nonTerminal.toObject().toString());
            }
        }
    }

    /**
     * This static method validates that the specified context-free grammar does
     * not contain left-recursive non-terminal symbols.
     *
     * @param grammar
     *            The context-free grammar to be validated.
     * @param errors
     *            The list to which errors that may occur can be added.
     */
    private static void validateIsLeftRecursionFree(final Grammar grammar, final List<String> errors) {
        for (final NonTerminal nonTerminal : Utils.computeLeftRecursiveNonTerminals(grammar)) {
            errors.add(nonTerminal.toObject().toString());
        }
    }

    /**
     * This static method validates that the specified context-free grammar does
     * not contain unit production rules.
     *
     * @param grammar
     *            The context-free grammar to be validated.
     * @param errors
     *            The list to which errors that may occur can be added.
     */
    private static void validateIsUnitFree(final Grammar grammar, final List<String> errors) {
        for (final Rule rule : grammar.getRules()) {
            if ((rule.getRightHandSide().size() == 1) && rule.getRightHandSide().get(0).isNonTerminal()) {
                errors.add(rule.toObject().toString());
            }
        }
    }

    /**
     * This constructor instantiates a new validator.
     */
    private Validator() {
    }
}
